﻿--[[	*** DataStore_Containers ***
Written by : Thaoky, EU-Marécages de Zangar
June 21st, 2009
--]]
DataStore_Containers = LibStub("AceAddon-3.0"):NewAddon("DataStore_Containers", "AceConsole-3.0", "AceEvent-3.0")

local addon = DataStore_Containers
local BI = LibStub("LibBabble-Inventory-3.0"):GetLookupTable()

local THIS_ACCOUNT = "Default"

local AddonDB_Defaults = {
	global = {
		['*'] = {							-- Account: "Default" or user defined for other accounts
			['*'] = {						-- Realm
				Guilds = {
					['*'] = {				-- guild["MyUberGuild"]
						['*'] = {		-- tabID = table index [1] to [6]
							name = nil,
							visitedBy = "",
							ClientTime = 0,				-- since epoch
							ClientDate = nil,
							ClientHour = nil,
							ClientMinute = nil,
							ServerHour = nil,
							ServerMinute = nil,
							ids = { ['*'] = nil },
							links = { ['*'] = nil },
							counts = { ['*'] = nil }
						}
					}
				},
				Characters = {
					['*'] = {					-- Character Name
 						numBagSlots = 0,
						numFreeBagSlots = 0,
						numBankSlots = 0,
						numFreeBankSlots = 0,
						Containers = {
							['*'] = {					-- Containers["Bag0"]
								icon = nil,				-- Containers's texture
								link = nil,				-- Containers's itemlink
								size = 0,
								freeslots = 0,
								bagtype = 0,
								ids = { ['*'] = nil },
								links = { ['*'] = nil },
								counts = { ['*'] = nil },
								cooldowns = { ['*'] = nil }
							}
						}
					}
				}
			}
		}
	}
}

function addon:OnInitialize()
	addon.db = LibStub("AceDB-3.0"):New("DataStore_ContainersDB", AddonDB_Defaults)
end

-- *** Utility functions ***
local function GetIDFromLink(link)
	if link then
		return tonumber(link:match("item:(%d+)"))
	end
end

local function IsEnchanted(link)
	if not link then return end
	
	if not string.find(link, "0:0:0:0:0:0:0") then
		-- enchants/jewels store values instead of zeroes in the link, if this string can't be found, there's at least one enchant/jewel
		return true
	end
end

local BAGS			= 1		-- All bags, 0 to 11, and keyring ( id -2 )
local BANK			= 2		-- 28 main slots
local GUILDBANK	= 3		-- 98 main slots

local ContainerTypes = {
	[BAGS] = {
		GetSize = function(self, bagID)
				return GetContainerNumSlots(bagID)
			end,
		GetFreeSlots = function(self, bagID)
				local freeSlots, bagType = GetContainerNumFreeSlots(bagID)
				return freeSlots, bagType
			end,
		GetLink = function(self, slotID, bagID)
				return GetContainerItemLink(bagID, slotID)
			end,
		GetCount = function(self, slotID, bagID)
				local _, count = GetContainerItemInfo(bagID, slotID)
				return count
			end,
		GetCooldown = function(self, slotID, bagID)
				local startTime, duration, isEnabled = GetContainerItemCooldown(bagID, slotID)
				return startTime, duration, isEnabled
			end,
	},
	[BANK] = {
		GetSize = function(self)
				return NUM_BANKGENERIC_SLOTS		-- 28
			end,
		GetFreeSlots = function(self)
				local freeSlots, bagType = GetContainerNumFreeSlots(-1)		-- -1 = player bank
				return freeSlots, bagType
			end,
		GetLink = function(self, slotID)
				return GetInventoryItemLink("player", slotID)
			end,
		GetCount = function(self, slotID)
				return GetInventoryItemCount("player", slotID)
			end,
		GetCooldown = function(self, slotID)
				local startTime, duration, isEnabled = GetInventoryItemCooldown("player", slotID)
				return startTime, duration, isEnabled
			end,
	},
	[GUILDBANK] = {
		GetSize = function(self)
				return MAX_GUILDBANK_SLOTS_PER_TAB		-- 98
			end,
		GetFreeSlots = function(self)
				return nil, nil
			end,
		GetLink = function(self, slotID, tabID)
				return GetGuildBankItemLink(tabID, slotID)
			end,
		GetCount = function(self, slotID, tabID)
				local _, count = GetGuildBankItemInfo(tabID, slotID)
				return count
			end,
		GetCooldown = function(self, slotID)
				return nil
			end,
	}
}

-- *** Scanning functions ***
local function ScanContainer(bagID, containerType)
	local Container = ContainerTypes[containerType]
	
	local bag
	if containerType == GUILDBANK then
		local guildName = GetGuildInfo("player")
		if not guildName then return end		-- to do, add an error message here, should not happen, since a scan is triggered when opening the guild bank
		
		bag = addon.ThisRealm.Guilds[GetGuildInfo("player")][bagID]	-- bag is actually the current tab
	else
		bag = addon.ThisCharacter.Containers["Bag" .. bagID]
		wipe(bag.cooldowns)		-- does not exist for a guild bank
	end

	wipe(bag.ids)				-- clean existing bag data
	wipe(bag.counts)
	wipe(bag.links)
	
	local link, count
	local startTime, duration, isEnabled
	
	bag.size = Container:GetSize(bagID)
	bag.freeslots, bag.bagtype = Container:GetFreeSlots(bagID)
	
	-- Scan from 1 to bagsize for normal bags or guild bank tabs, but from 40 to 67 for main bank slots
	local baseIndex = (containerType == BANK) and 39 or 0
	local index
	
	for slotID = baseIndex + 1, baseIndex + bag.size do
		index = slotID - baseIndex
		link = Container:GetLink(slotID, bagID)
		if link then
			bag.ids[index] = GetIDFromLink(link)

			if IsEnchanted(link) then
				bag.links[index] = link
			end
		
			count = Container:GetCount(slotID, bagID)
			if count and count > 1  then
				bag.counts[index] = count	-- only save the count if it's > 1 (to save some space since a count of 1 is extremely redundant)
			end
		end
		
		startTime, duration, isEnabled = Container:GetCooldown(slotID, bagID)
		if startTime and startTime > 0 then
			bag.cooldowns[index] = startTime .."|".. duration .. "|" .. 1
		end
	end
end

local function ScanBagSlotsInfo()
	local char = addon.ThisCharacter

	local numBagSlots = 0
	local numFreeBagSlots = 0

	for bagID = 0, NUM_BAG_SLOTS do
		local bag = char.Containers["Bag" .. bagID]
		numBagSlots = numBagSlots + bag.size
		numFreeBagSlots = numFreeBagSlots + bag.freeslots
	end
	
	char.numBagSlots = numBagSlots
	char.numFreeBagSlots = numFreeBagSlots
end

local function ScanBankSlotsInfo()
	local char = addon.ThisCharacter
	
	local numBankSlots = NUM_BANKGENERIC_SLOTS
	local numFreeBankSlots = char.Containers["Bag100"].freeslots

	for bagID = NUM_BAG_SLOTS, NUM_BAG_SLOTS + NUM_BANKBAGSLOTS do		-- 5 to 11
		local bag = char.Containers["Bag" .. bagID]
		
		numBankSlots = numBankSlots + bag.size
		numFreeBankSlots = numFreeBankSlots + bag.freeslots
	end
	
	char.numBankSlots = numBankSlots
	char.numFreeBankSlots = numFreeBankSlots
end

local function ScanGuildBankInfo()
	-- only the current tab can be updated
	local guildName = GetGuildInfo("player")
	local tabID = GetCurrentGuildBankTab()
	local t = addon.ThisRealm.Guilds[guildName][tabID]	-- t = current tab

	t.name = GetGuildBankTabInfo(tabID)
	t.visitedBy = UnitName("player")
	t.ClientTime = time()
	if GetLocale() == "enUS" then				-- adjust this test if there's demand
		t.ClientDate = date("%m/%d/%Y")
	else
		t.ClientDate = date("%d/%m/%Y")
	end
	t.ClientHour = tonumber(date("%H"))
	t.ClientMinute = tonumber(date("%M"))
	t.ServerHour, t.ServerMinute = GetGameTime()
end

local function ScanBag(bagID)
	if bagID < 0 then return end

	local char = addon.ThisCharacter
	local bag = char.Containers["Bag" .. bagID]
	
	if bagID == 0 then	-- Bag 0	
		bag.icon = "Interface\\Buttons\\Button-Backpack-Up";
		bag.link = nil;
	else						-- Bags 1 through 11
		bag.icon = GetInventoryItemTexture("player", ContainerIDToInventoryID(bagID))
		bag.link = GetInventoryItemLink("player", ContainerIDToInventoryID(bagID))
	end
	ScanContainer(bagID, BAGS)
	ScanBagSlotsInfo()
end

local function ScanKeyRing()
	local char = addon.ThisCharacter
	local bag = char.Containers["Bag" .. KEYRING_CONTAINER]

	bag.icon = "Interface\\Icons\\INV_Misc_Key_14";
	bag.link = nil
	ScanContainer(KEYRING_CONTAINER, BAGS)
end

function addon:OnEnable()
	local realm = GetRealmName()
	local player = UnitName("player")
	
	self.ThisAccount = addon.db.global[THIS_ACCOUNT]
	self.ThisRealm = self.ThisAccount[realm]
	self.ThisCharacter = self.ThisRealm.Characters[player]

	-- manually update bags 0 to 4, then register the event, this avoids reacting to the flood of BAG_UPDATE events at login
	for bagID = 0, NUM_BAG_SLOTS do
		ScanBag(bagID)
	end
	ScanKeyRing()
	
	self:RegisterEvent("BAG_UPDATE")
	self:RegisterEvent("BANKFRAME_OPENED")
	self:RegisterEvent("GUILDBANKFRAME_OPENED")
end

function addon:OnDisable()
	self:UnregisterEvent("BAG_UPDATE")
	self:UnregisterEvent("BANKFRAME_OPENED")
	self:UnregisterEvent("GUILDBANKFRAME_OPENED")
end


-- *** DB functions ***
function addon:GetCharacter(name, realm, account)
	-- assert(type(name) == "string")
	-- assert(type(realm) == "string")
	-- assert(type(account) == "string")
	
	-- Usage: 
	-- 	local c = addon:GetCharacter(char, realm, account)
	--	all 3 parameters default to current player, realm or account
	name = name or UnitName("player")
	realm = realm or GetRealmName()
	account = account or THIS_ACCOUNT
	
	return addon.db.global[account][realm].Characters[name]
end

function addon:DeleteCharacter(name, realm, account)
	name = name or UnitName("player")
	realm = realm or GetRealmName()
	account = account or THIS_ACCOUNT

	addon.db.global[account][realm].Characters[name] = nil
end

function addon:GetGuild(name, realm, account)
	-- assert(type(name) == "string")
	-- assert(type(realm) == "string")
	-- assert(type(account) == "string")
	
	-- Usage: 
	-- 	local c = addon:GetCharacter(char, realm, account)
	--	all 3 parameters default to current player, realm or account
	realm = realm or GetRealmName()
	account = account or THIS_ACCOUNT
	
	return addon.db.global[account][realm].Guilds[name]
end

function addon:DeleteGuild(name, realm, account)
	realm = realm or GetRealmName()
	account = account or THIS_ACCOUNT
	
	addon.db.global[account][realm].Guilds[name] = nil
end

function addon:GetContainer(character, containerID)
	assert(type(character) == "table")		-- this is the pointer to a character table, obtained through addon:GetCharacter()
	-- containerID can be number or string
	
	return character.Containers["Bag" .. containerID]
end

function addon:GetContainers(character)
	assert(type(character) == "table")		-- this is the pointer to a character table, obtained through addon:GetCharacter()
		
	return character.Containers
end

local BagTypeStrings = {
	[1] = BI["Quiver"],
	[2] = BI["Ammo Pouch"],
	[4] = BI["Soul Bag"],
	[8] = BI["Leatherworking Bag"],
	[16] = BI["Inscription Bag"],
	[32] = BI["Herb Bag"],
	[64] = BI["Enchanting Bag"],
	[128] = BI["Engineering Bag"],
	[512] = BI["Gem Bag"],
	[1024] = BI["Mining Bag"],
}

function addon:GetContainerInfo(character, containerID)
	assert(type(character) == "table")		-- this is the pointer to a character table, obtained through addon:GetCharacter()
	-- containerID can be number or string
	
	local bag = character.Containers["Bag" .. containerID]
	return bag.icon, bag.link, bag.size, bag.freeslots, (BagTypeStrings[bag.bagtype] or "")
end

function addon:GetContainerSize(source, containerID)
	-- containerID can be number or string
	-- two types of calls are accepted : 
		-- addon:GetContainerSize(source, containerID), where source is a character table, as returned by addon:GetCharacter()
		-- addon:GetContainerSize(source), where source is a container table, as returned when iterating through containers with addon:GetContainers(character)

	assert(type(source) == "table")
	
	if source.size then		-- if a container table was passed, return its size directly
		return source.size
	end

	return source.Containers["Bag" .. containerID].size
end

function addon:GetSlotInfo(bag, slotID)
	assert(type(bag) == "table")		-- this is the pointer to a bag table, obtained through addon:GetContainer()
	assert(type(slotID) == "number")

	-- return itemID, itemLink, itemCount
	return bag.ids[slotID], bag.links[slotID], bag.counts[slotID]
end

function addon:GetCooldownInfo(bag, slotID)
	assert(type(bag) == "table")		-- this is the pointer to a bag table, obtained through addon:GetContainer()
	assert(type(slotID) == "number")

	local cd = bag.cooldowns[slotID]
	if cd then
		local startTime, duration, isEnabled = strsplit("|", bag.cooldowns[slotID])
		local remaining = duration - (GetTime() - startTime)
		
		if remaining > 0 then		-- valid cd ? return it
			return tonumber(startTime), tonumber(duration), tonumber(isEnabled)
		end
		-- cooldown expired ? clean it from the db
		bag.cooldowns[slotID] = nil
	end
end

function addon:GetItemCount(character, searchedID)
	assert(type(character) == "table")		-- this is the pointer to a character table, obtained through addon:GetCharacter()

	local bagCount = 0
	local bankCount = 0
	local id
	
	for containerName, container in pairs(character.Containers) do
		for slotID=1, container.size do
			id = container.ids[slotID]
			
			if (id) and (id == searchedID) then
				local itemCount = container.counts[slotID] or 1
				
				if (containerName == "Bag100") then
					bankCount = bankCount + itemCount
				elseif (containerName == "Bag-2") then
					bagCount = bagCount + itemCount
				else
					local bagNum = tonumber(string.sub(containerName, 4))
					if (bagNum >= 0) and (bagNum <= 4) then
						bagCount = bagCount + itemCount
					else
						bankCount = bankCount + itemCount
					end			
				end		
			end
		end
	end

	return bagCount, bankCount
end

function addon:GetGuildBankTab(guild, tabID)
	assert(type(guild) == "table")		-- this is the pointer to a character table, obtained through addon:GetGuild()
	
	return guild[tabID]
end

function addon:GetGuildBankTabName(guild, tabID)
	assert(type(guild) == "table")		-- this is the pointer to a character table, obtained through addon:GetGuild()
	
	return guild[tabID].name
end

-- *** EVENT HANDLERS ***
function addon:BAG_UPDATE(event, bag)
	if bag < 0 then
		return
	end
	
	if (bag >= 5) and (bag <= 11) and not addon.isBankOpen then
		return
	end
	
	if bag == 0 then					-- bag is 0 for both the keyring and the original backpack
		ScanKeyRing()
	end
	ScanBag(bag)
end

function addon:BANKFRAME_OPENED()
	addon.isBankOpen = true
	for bagID = NUM_BAG_SLOTS, NUM_BAG_SLOTS + NUM_BANKBAGSLOTS do		-- 5 to 11
		ScanBag(bagID)
	end
	ScanContainer(100, BANK)
	ScanBankSlotsInfo()
	addon:RegisterEvent("BANKFRAME_CLOSED")
	addon:RegisterEvent("PLAYERBANKSLOTS_CHANGED")
end

function addon:BANKFRAME_CLOSED()
	addon.isBankOpen = nil
	addon:UnregisterEvent("BANKFRAME_CLOSED")
	addon:UnregisterEvent("PLAYERBANKSLOTS_CHANGED")
end

function addon:PLAYERBANKSLOTS_CHANGED(event, slotID)
	-- from top left to bottom right, slotID = 1 to 28for main slots, and 29 to 35 for the additional bags
	if (slotID >= 29) and (slotID <= 35) then
		ScanBag(slotID - 24)		-- bagID for bank bags goes from 5 to 11, so slotID - 24
	else
		ScanContainer(100, BANK)
		ScanBankSlotsInfo()
	end
end

function addon:GUILDBANKFRAME_OPENED()
	addon:RegisterEvent("GUILDBANKFRAME_CLOSED")
	addon:RegisterEvent("GUILDBANKBAGSLOTS_CHANGED")
end

function addon:GUILDBANKBAGSLOTS_CHANGED()
	ScanContainer(GetCurrentGuildBankTab(), GUILDBANK)
	ScanGuildBankInfo()
end

function addon:GUILDBANKFRAME_CLOSED()
	addon:UnregisterEvent("GUILDBANKFRAME_CLOSED")
	addon:UnregisterEvent("GUILDBANKBAGSLOTS_CHANGED")
end
